modules = [

    #"logic",

    "decode",
    "decoder",
    "encoder",
    "arbiter",

    'halfadder',
    'fulladder',
    "arith",

    "compare",

    "barrel",

    "register",

    "siso",
    "sipo",
    "piso",
    "pipo",

    "ring",
    "cascadedring",

    "johnson",

    "counter",
    "countermod",
    "counterload",

    "RAM",
    "ROM",

    "primitives",
    "operator",

    "reg_file",
]
